/*********************

Thermostat for Space Heater - sh_therm_1-3

version 1.3

Read temperature sensor.
Average a number of readings (TEMP_BUF_SIZE readings) to minimumize noise
Turn space heater on or off via 233MHz remote (hacked from remote control mains outlets

Controls:
Set Temp
On/Off

Display:
Room Temperature
Set Point
On/Off State

Started from Adifruit I2C RGB LCD Shield Hello World sketch

to do...

1. I'm not sure I have the EEPROM wear leveling library implimented correctly. I ended up using
EEPROM.put() and EEPROM.get() while initializing with EEPROMwl.begin(). When I changed
EEPROM to EEPROMwl it quit saving to EEPROM. I don't get it but I'm going to ignore it for now.

2. Maybe someday - throw away completely invalid sensor values in checkTemp
  (though I've not seen any yet). 



**********************/

// include the library code:
#include <Wire.h>
#include <Adafruit_RGBLCDShield.h>
#include <utility/Adafruit_MCP23017.h>
#include <math.h>
#include "Timer.h"
#include <EEPROMWearLevel.h>


// The LCD shield uses the I2C SCL and SDA pins.

Adafruit_RGBLCDShield lcd = Adafruit_RGBLCDShield();


//************** Defines **********************

#define TEMP_CK_TIME  1000        //check timer every n ms
#define WHITE 0x7                 //LCD backlight color
#define BLUE 0x4                  //turns the backlight off on the lcd I'm using
#define DISPLAY_MS 12000          //turn lcd off after xxx ms
#define TEMP_BUF_SIZE 5           //number of temperature readings to average
#define INC_CHANGE 1              //amount of change to allow if reading was noise
#define INIT_TEMP  (double)68.0           //set temp at first startup (use eeprom later) - must be float
#define HYSTERESIS 1              //difference between turn on temp and turn off temp
#define SIGNAL_TIME 50            //time to hold remote line low in ms

// io defines
#define AD_PIN A0                 //a to d pin connected to temp sensor
#define HEATER_ON_PIN 8
#define HEATER_OFF_PIN 7

// eeprom defines
#define EEPROM_LAYOUT_VERSION 0
#define AMOUNT_OF_INDEXES 1
#define INDEX_CONFIGURATION_VAR1 0
#define INDEX_CONFIGURATION_VAR2 1

#define EEPROM_WRITE_DELAY 5000     //delay after temperature setting is changed before writing to EEPROM to prevent extra writes



//***************Globals ***********************

Timer t;

int timerID;
int updateStuffTimerID;
unsigned long LCDmillisTarget = 10000;
unsigned long eepromMillisTarget = 0;

double temperature;
double lastTemp = 11;
double setTemp;

double tempArray[TEMP_BUF_SIZE];
char buffer[16];
uint8_t buttons;
bool onOff = 0;
bool heaterOn = 0;

//eeprom 
int eeAddress = 0; 



//************** Setup *************************

void setup() {
  // Debugging output
  Serial.begin(9600);
  //while (!Serial) {
    //; // wait for serial port to connect. Needed for native USB port only
  //}

  //set up io
  pinMode(HEATER_ON_PIN, OUTPUT);             // set pin to output
  digitalWrite(HEATER_ON_PIN, HIGH);
  pinMode(HEATER_OFF_PIN, OUTPUT);            // set pin to output 
  digitalWrite(HEATER_OFF_PIN, HIGH);

  //turn heater off
  TurnHeaterOff();
    
  // set up the LCD's number of columns and rows: 
  lcd.begin(16, 2);
  
  //print splash screen for a second
  lcd.setBacklight(WHITE);
  lcd.setCursor(2, 0);
  lcd.print("Space Heater");
  lcd.setCursor(3, 1);
  lcd.print("Thermostat");
  //delay(1000);
  
  initTemp();       //set up temperature array with valid values
  
  //set up screen for running
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Temperature");

  lcd.setCursor(5, 1);
  lcd.print("Set To");

  lcd.setCursor(13,1);
  lcd.print("OFF");

  timerID = t.every(TEMP_CK_TIME, checkTemp);      //set up timer to call checkTemp regularly

  //create an event timer to write eeprom and turn of lcd
  updateStuffTimerID = t.every(1000, updateStuff);
  
  //load eeprom value for setTemp
  //eeprom setup stuff
  //first, set up eeprom
  EEPROMwl.begin(EEPROM_LAYOUT_VERSION, AMOUNT_OF_INDEXES);
  
  //get setTemp from eeprom
  EEPROM.get(eeAddress, setTemp);                 //get value from eeprom

  //is it a number? if not, this is our first time running
  if(isnan(setTemp)){
    //it's not a number, use initial define
    //Serial.println("not a number"); 
    setTemp = INIT_TEMP; 
    //write to eeprom
    EEPROM.put(eeAddress, setTemp); 
    
  }else{
    //it's a number, is it a valid one?
    //Serial.println("was a number");

    double foo = abs(setTemp - INIT_TEMP);
    //Serial.print(foo, 4);

    if(foo > (double)40.0){
      //wasn't a valid number
      //Serial.println(" wasn't valid");
      
      //write default to eeprom
      setTemp = INIT_TEMP; 
      //write to eeprom
      EEPROM.put(eeAddress, setTemp);
      
    }
  }
}

//********************************* main loop **************************************
void loop() {

  t.update();

  //Serial.println(temperature);

  //check buttons
 
  //------ ANY BUTTON?--------- 
  buttons = lcd.readButtons();

  if((buttons > 0) && (LCDmillisTarget == 0)){       //----- BUTTON PRESSED AND DISPLAY OFF -----
    //if button pressed and display was off,
    // turn on lcd display
    
    //Serial.println(buttons);
    lcd.display();
    lcd.setBacklight(WHITE);
    LCDmillisTarget = millis() + DISPLAY_MS;

    //wait for button release
    while(buttons){
      buttons = lcd.readButtons();
    }
    
  } else if (buttons & BUTTON_SELECT){               //----- ON / OFF BUTTON ---------

    //toggle on/off state
    onOff ^= 1;

    if(onOff){
      //turn on
      //Serial.println("turned on");
      sprintf(buffer,"%3d",(int)(setTemp + 0.5));
      lcd.setCursor(13,1);
      lcd.print(buffer);
     }else{
      //turn off
      heaterOn = 0;
      
      //turn off heater
      TurnHeaterOff();


      //erase * symbol
      lcd.setCursor(12,1);
      lcd.print(" ");

      //display word "off" on LCD
      lcd.setCursor(13,1);
      lcd.print("OFF");
    }

    LCDmillisTarget = millis() + DISPLAY_MS;

    //wait for button release
    while(buttons){
      buttons = lcd.readButtons();
    }

  
  }else if((buttons & BUTTON_UP) && onOff){               //-------- UP BUTTON ----------  
    //up button pressed and it's on, increment setTemp
    setTemp ++;
    sprintf(buffer,"%3d",(int)(setTemp + 0.5));
    lcd.setCursor(13,1);
    lcd.print(buffer);

    LCDmillisTarget = millis() + DISPLAY_MS;
   
    //wait for button release
    while(buttons){
      buttons = lcd.readButtons();
    }
 
    //set timer for eeprom write
    eepromMillisTarget = millis() + EEPROM_WRITE_DELAY;

  
  }else if ((buttons & BUTTON_DOWN) && onOff){            //-------- DOWN BUTTON ------------ 
    //down button pushed and it's on, decrement setTemp
    setTemp --;
    sprintf(buffer,"%3d",(int)(setTemp + 0.5));
    lcd.setCursor(13,1);
    lcd.print(buffer);

    LCDmillisTarget = millis() + DISPLAY_MS;

    //wait for button release
    while(buttons){
      buttons = lcd.readButtons();
    }

    //setup for eeprom write
    eepromMillisTarget = millis() + EEPROM_WRITE_DELAY;
    
  }

  if(onOff && !heaterOn && (temperature < (setTemp - HYSTERESIS))){
    //if unit is on (onOff == 1) 
    //and setTemp > temperature, 
    //and we're not already on
    
    //display * symbol
    
      lcd.setCursor(12,1);
      lcd.print("*");
      
    //turn on heat
      heaterOn = 1;
      TurnHeaterOn();

  }

  if(onOff && heaterOn && (temperature >= setTemp)){
    //if unit is on
    //and setTemp <= temperature
    //and we're not already off

    //erase * symbol
    lcd.setCursor(12,1);
    lcd.print(" ");

    //turn heat off
    heaterOn = 0;
    TurnHeaterOff();
  }
  
  //if temp changed, display it
  if(lastTemp != temperature){
  
    sprintf(buffer,"%3d",(int)(temperature + 0.5));
    lcd.setCursor(13,0);
    lcd.print(buffer);

    lastTemp = temperature;

  }

}

//************** Functions *********************


//******** TurnHeaterOff() *********
void TurnHeaterOff(){
  digitalWrite(HEATER_OFF_PIN, LOW);
  delay(SIGNAL_TIME);
  digitalWrite(HEATER_OFF_PIN, HIGH);  
}


//******** TurnHeaterOn() *********
void TurnHeaterOn(){
  digitalWrite(HEATER_ON_PIN, LOW);
  delay(SIGNAL_TIME);
  digitalWrite(HEATER_ON_PIN, HIGH);  
}


//******* initTemp ******
void initTemp(){
  //load up the temperature circular buffer
  int i;
  double temp;
  double sum;

  analogRead(AD_PIN);
  delay(TEMP_CK_TIME/2);
  
  for (i = 0; i < TEMP_BUF_SIZE; ++i){

    temp = log(((10240000/analogRead(AD_PIN)) - 10000));
    temp = 1 / (0.001129148 + (0.000234125 + (0.0000000876741 * temp * temp ))* temp );
    temp = temp - 273.15; // Convert Kelvin to Celcius
    temp = (temp * 1.8) + 32;         //convert to fahrenheit for those of us that can't manage to think in C
    
    tempArray[i] = temp;
    //Serial.println(i);
    //Serial.println(temp);
    delay(TEMP_CK_TIME/2);
    
  }

  for(i = 0; i < TEMP_BUF_SIZE; i++){
    sum = sum + tempArray[i];   
  }
  temperature = sum / TEMP_BUF_SIZE;

  //Serial.println("Average");
  //Serial.println(temperature);

}


//******* check temp *******
void checkTemp(){
   //check the temperature sensor
   //called by timer function every xxx ms

  double temp;
  double sum;
  int i;
  double diff;
  
  temp = log(((10240000/analogRead(AD_PIN)) - 10000));
  temp = 1 / (0.001129148 + (0.000234125 + (0.0000000876741 * temp * temp ))* temp );
  temp = temp - 273.15; // Convert Kelvin to Celcius
  //convert to fahrenheit for those of us that can't manage to think in C
  temp = (temp * 1.8) + 32;

  diff = temperature - temp;

  //put it in tempArray if it's a valid reading, if not ignore
  //shuffle array left
  for (i = TEMP_BUF_SIZE; i > 1; i--){
    //Serial.println(i);
    tempArray[i-2] = tempArray[i-1];
   }
   
  tempArray[TEMP_BUF_SIZE - 1] = temp;

  //return average of array in temperature

  for(i = 0; i < TEMP_BUF_SIZE; ++i){
    sum = sum + tempArray[i];   
  }
  temperature = sum / TEMP_BUF_SIZE;
}


//******* updateStuff() ******
void updateStuff(){
  //called by timer function every second or so
  unsigned long ms;

  ms = millis();
  
  //see if eeprom needs writing
  if((eepromMillisTarget != 0) && (ms > eepromMillisTarget)){
    EEPROM.put(eeAddress, setTemp); 
    eepromMillisTarget = 0;
   
  }

  //see if display needs to be turned off
  if((LCDmillisTarget != 0) && (ms > LCDmillisTarget)){
    lcd.noDisplay();
    lcd.setBacklight(BLUE);
    LCDmillisTarget = 0;
  }
}




